using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Ants
{
    public class Ant
    {
        
        //Fixed parameters for ant movement 
        public float Weight = (float)(2.5 * Math.Pow(10, -6));
        public const float RestingMR = 4.1F;
        const float Kappa = 0.43F;
        const float MaxKappa = Kappa * 3F;
        const float MaxEnergy = 24F;
        const float ReturnHomeFactor = MaxEnergy * -1;
        const float FeedScale = 100F;
        const float DepositRate = 0.0033F;
        const float MaxAngleChange = (float)(Math.PI / 4);
        const float InitialPheromoneAmount = 1F;
        const int MaximumPathLength = 100000;

        float fEnergy;
        float fDistance;
        float fDirection;
        float fMeanDistance;
        float fMaxDistance;
        float fStandardDevDistance;

        bool fRouteFidelity;

        bool fFoundFood;
        bool fFoundNest;
        bool fLayTrail;
        bool fAtTarget;

        bool fstealbool;
        bool ffeedbool;
        bool fFoundFoodEmpty;
        bool fFoundNestEmpty;

        PointF fPosition;
        PointF fTarget;
        Nest fHomeNest;
        Nest fStealNest;
        Colony fColony;
        Behaviour fBehaviour;
        World fWorld;
        TargetType fTargetType;
        AntType fAntType;
        Pen fPen = new Pen(Color.Black);
        List<Nest> fVisitedNests = new List<Nest>();
        List<Food> fVisitedFoods = new List<Food>();

        public Dictionary<Nest, int> fAllVisitedNests = new Dictionary<Nest, int>();
        public Dictionary<Food, int> fAllVisitedFoods = new Dictionary<Food, int>();

        PointF AntennaLeft
        {
            get
            {
                PointF head = fHeadPosition;
                double angle = fDirection - (Math.PI / 4.0);

                return new PointF(head.X + Pheromone.SensingDistance * ((float)Math.Sin(angle)),
                    head.Y - Pheromone.SensingDistance * ((float)Math.Cos(angle)));
            }
        }

        PointF AntennaRight
        {
            get
            {
                PointF head = fPosition;
                double angle = fDirection + (Math.PI / 4.0);

                return new PointF(head.X + Pheromone.SensingDistance * ((float)Math.Sin(angle)),
                    head.Y - Pheromone.SensingDistance * ((float)Math.Cos(angle)));
            }
        }

        PointF fHeadPosition
        {
            get
            {
                return new PointF(fPosition.X + Ant.Length * ((float)Math.Sin(fDirection)),
                    fPosition.Y - Ant.Length * ((float)Math.Cos(fDirection)));
            }
        }

        public enum AntType
        {
            Forager = 0,
            Transporter = 1,
        }

        enum Behaviour
        {
            Dead = 0,
            InNest = 1,
            Foraging = 2,
            Feeding = 3,
            TakingFromNest = 4,
            GoingToTarget = 5,
            Following = 6,
            Waiting = 7
        }

        enum TargetType
        {
            NoTarget = 0,
            Nest = 1,
            Food = 2,
            Home = 3
        }

        public const float Length = 4F;
        public float Direction
        {
            get
            {
                return fDirection;
            }
            set
            {
                if ((value != Double.NaN) && !Double.IsInfinity(value))
                {
                    fDirection = value;

                    while (fDirection < 0)
                        fDirection += (float)(2 * Math.PI);

                    while (fDirection > 2 * Math.PI)
                        fDirection -= (float)(2 * Math.PI);
                }
            }
        }
        public PointF Position
        {
            get { return fPosition; }
        }
        public float MeanDistance
        {
            get { return fMeanDistance; }
        }
        public float StandardDevDistance
        {
            get { return fStandardDevDistance; }
        }
        public List<int> fWaitingTimes = new List<int>();

        public int  fStepAfterLeaving = 0;

        //Constructor
        public Ant(Nest aNest, bool aRouteFidelity, AntType aAntType, float aSpeed = 4.2F, float aStDevSpeed = 2F)
        {
            fHomeNest = aNest;
            fMeanDistance = aSpeed;
            fMaxDistance = 5 * aSpeed;
            fStandardDevDistance = aStDevSpeed;
            fBehaviour = Behaviour.Foraging;
            fVisitedNests.Add(fHomeNest);
            fRouteFidelity = aRouteFidelity;
            fstealbool = true;
            ffeedbool = true;
            fAntType = aAntType;

            fPosition = new PointF(fHomeNest.Position.X, fHomeNest.Position.Y);
            fDirection = 2 * (float)Math.PI * (float)fHomeNest.Random;
            fFoundFood = false;
            fLayTrail = false;
            fEnergy = 0;
            fTargetType = TargetType.NoTarget;
            fWorld = fHomeNest.World;
            fColony = aNest.Colony;
        }

        /// <summary>
        /// Checks type and mode of ant and carries out appropriate action
        /// </summary>
        public void Step()
        {
            switch (fBehaviour)
            {
                case Behaviour.InNest:
                    {
                        ResetAnt();
                        //Foragers exchange food with nest or leave nest
                        if (!NestExchange())
                        {
                            LeaveHomeNest();
                        }
                    }
                    break;

                case Behaviour.Foraging:
                    {
                        //Ant will stop foraging and head home with probability dependent on energy 
                        if ((fHomeNest.Random * ReturnHomeFactor) <= fEnergy)
                        {
                            //if ignoring pheromones near nest need to check how many steps from nest
                            if (fWorld.IgnoreCloseToNest)
                            {
                                if (fStepAfterLeaving <= 5)
                                {
                                    MoveRandom();
                                }
                                else
                                {
                                    PheromoneMove();
                                }
                            }
                            else
                            {
                                PheromoneMove();
                            }

                            fStepAfterLeaving++;
                        }
                        else
                        {
                            SetOffHome();
                        }
                    }
                    break;

                case Behaviour.GoingToTarget:
                    {
                        MoveToTarget();
                    }
                    break;

                case Behaviour.Feeding:
                    {
                        fVisitedNests.Clear();
                        fVisitedFoods.Clear();
                        Food FoodSource = fHomeNest.World.FindFoodAt(fPosition, Food.fRadius);
                        if (ffeedbool)
                        {
                            if (FoodSource != null)
                            {
                                if (fAllVisitedFoods.ContainsKey(FoodSource))
                                {
                                    fAllVisitedFoods[FoodSource] = fAllVisitedFoods[FoodSource] + 1;
                                    ffeedbool = false;
                                }
                            }
                        }
                        //Ant will stay feeding if food still exists with probability dependent on energy
                        if (!TakeFood(FoodSource))
                        {
                            ffeedbool = true;
                            SetOffHome();
                        }

                        fDistance = 0;
                    }
                    break;

                case Behaviour.TakingFromNest:
                    {
                        fVisitedNests.Clear();
                        fVisitedFoods.Clear();
                        fStealNest = fHomeNest.World.FindNestAt(fPosition, Nest.fRadius);
                        fVisitedNests.Add(fStealNest);
                        if (fstealbool)
                        {
                            fAllVisitedNests[fStealNest] = fAllVisitedNests[fStealNest] + 1;
                            fstealbool = false;
                        }
                        
                        //Ant will stay stealing if nest still has food with probability dependent on energy
                        if (!StealFood())
                        {
                            fstealbool = true;
                            SetOffHome();
                        }
                        fDistance = 0;
                    }
                    break;
            }

            //Ant always uses energy according to the distance it has travelled
            fEnergy -= (RestingMR + (0.75F * (fDistance))) * Weight;
        }

        //Reset for when ant returns to nest
        void ResetAnt()
        {
            this.fVisitedNests.Clear();
            this.fVisitedFoods.Clear();
            this.fDistance = 0;
        }

        //Clears parameters and sets ant to foraging
        void LeaveHomeNest()
        {
            fVisitedNests.Add(fHomeNest);
            fBehaviour = Behaviour.Foraging;
            fTargetType = TargetType.NoTarget;

            if (fRouteFidelity && (fFoundFood || fFoundNest)) //Ants leave home in the direction they arrived
            {
                if (!(fFoundFoodEmpty || fFoundNestEmpty))
                {
                    fDirection = fDirection + (float)(Math.PI);
                }
            }
            else
            {
                fDirection = (float)(this.fWorld.Random * Math.PI * 2);
            }

            fAtTarget = false;
            fHomeNest.WaitingAnts.Remove(this);
            fFoundFood = false;
            fFoundNest = false;
            fLayTrail = false;
            fStealNest = null;
            fFoundFoodEmpty = false;
            fFoundNestEmpty = false;
            fStepAfterLeaving = 1;
        }

        //Moves ant according to correlated random walk
        void MoveRandom()
        {
            GetDistance();
            GetRandomDirection();
            UpdatePosition();
            CheckForTargets(Food.DetectRange);
        }

        //Moves ant towrds target with randomness
        void MoveToTarget()
        {
            //Check for being at (within detection range) target
            if (World.IsPointInRadius(fPosition, fTarget, Food.DetectRange))
            {
                fDistance = (float)(Math.Sqrt((fPosition.X - fTarget.X) * (fPosition.X - fTarget.X) + (fPosition.Y - fTarget.Y) * (fPosition.Y - fTarget.Y)));
                fPosition = fTarget;
                fAtTarget = true;
            }

            else
            {
                GetDistance();
                //If energy is < 0 and ant in on way home then ant checks for food
                if ((fEnergy <= 0) && (fBehaviour == Behaviour.GoingToTarget) && (fTargetType == TargetType.Home))
                {
                    CheckForTargets(Food.DetectRange);
                }

                if (fTargetType == TargetType.Home)
                {
                    if (fLayTrail)
                    {
                        fColony.AddPheromone(fPosition, InitialPheromoneAmount, Pheromone.Type.FoundFood);
                    }
                }

                GetTargetDirection(fTarget);
                UpdatePosition();
            }

            //Once at target change mode as appropriate
            if (fAtTarget)
            {
                if (fTargetType == TargetType.Food)
                {
                    fBehaviour = Behaviour.Feeding;
                }
                if (fTargetType == TargetType.Nest)
                {
                    //behaviour to taking from nest
                    fBehaviour = Behaviour.TakingFromNest;
                    //record ants arriving from home nest
                    Nest NestAt = fHomeNest.World.FindNestAt(fPosition, Nest.fRadius);
                    int value = NestAt.ArrivalsFromNests[fHomeNest];
                    value++;
                    NestAt.ArrivalsFromNests[fHomeNest] = value;
                }

                if (fTargetType == TargetType.Home)
                {
                    fBehaviour = Behaviour.InNest;
                    //record ants coming from another nest........
                    if (fFoundNest)
                    {
                        int value = fHomeNest.ArrivalsFromNests[fStealNest];
                        value++;
                        fHomeNest.ArrivalsFromNests[fStealNest] = value;
                    }
                }
            }
        }

        //Checks for pheromones and moves ant
        void PheromoneMove()
        {
            PointF checkLeft = this.AntennaLeft;
            PointF checkRight = this.AntennaRight;

            double leftWeight = 0;
            double rightWeight = 0;
            double pheromoneWeight = 1;

            double newDirection = fDirection;

            RectangleF leftRegion = new RectangleF(checkLeft.X - Pheromone.SensingDistance, checkLeft.Y - Pheromone.SensingDistance, Pheromone.SensingDistance * 2, Pheromone.SensingDistance * 2);
            RectangleF rightRegion = new RectangleF(checkRight.X - Pheromone.SensingDistance, checkRight.Y - Pheromone.SensingDistance, Pheromone.SensingDistance * 2, Pheromone.SensingDistance * 2);

            //Pheromone weighting for movement
            double leftPheromones = fColony.CheckPheromone(Pheromone.Type.FoundFood, leftRegion);
            double rightPheromones = fColony.CheckPheromone(Pheromone.Type.FoundFood, rightRegion);

            if (leftPheromones + rightPheromones > 0)
            {
                leftWeight += leftPheromones * pheromoneWeight;
                rightWeight += rightPheromones * pheromoneWeight;
            }

            //If no pheromone add random deviation in direction
            if (!(leftWeight + rightWeight > 0))
            {
                fDirection += GetVonMisesAngle(Kappa);
            }

            if (leftWeight + rightWeight != 0)
            {
                //calculates direction change by proportion on right 
                double angleChange = (rightWeight - leftWeight) / (rightWeight + leftWeight);

                //stop from truning more than MaxAngle
                if (angleChange < -1 * MaxAngleChange)
                    angleChange = -1 * MaxAngleChange;
                else if (angleChange > MaxAngleChange)
                    angleChange = MaxAngleChange;

                //The stronger the pheromones the less randomness in turning angle
                float Randomness = (float)(Kappa + (leftWeight + rightWeight) / 5);
                //Maximum deviation when following pherones is MaxKappa
                Randomness = Math.Min(Randomness, MaxKappa);
                //Add randomness to new direction
                fDirection += (float)((angleChange * MaxAngleChange) + GetVonMisesAngle(Randomness));
            }

            GetDistance();
            UpdatePosition();
            CheckForTargets(Food.DetectRange);
        }

        //Check if nests and food are within dectection range
        void CheckForTargets(float aRadius)
        {
            if (fAntType == AntType.Forager)
            {
                // Check for food
                if (CheckForFood(aRadius))
                {
                    Food FoundFood = fHomeNest.World.FindFoodAt(fPosition, aRadius);
                    if (!fVisitedFoods.Contains(FoundFood))
                    {
                        fTarget = new PointF(FoundFood.Position.X, FoundFood.Position.Y);
                        fTargetType = TargetType.Food;
                        fBehaviour = Behaviour.GoingToTarget;
                    }
                }
            }

            if (fAntType == AntType.Transporter)
            {
                if (CheckForNest(aRadius))
                {
                    Nest FoundNest = fHomeNest.World.FindNestAt(fPosition, aRadius);
                    if (!fVisitedNests.Contains(FoundNest))
                    {
                        fTarget = new PointF(FoundNest.Position.X, FoundNest.Position.Y);
                        fTargetType = TargetType.Nest;
                        fBehaviour = Behaviour.GoingToTarget;
                    }
                }    
            }
        }

        //Updates fDistance by drawing a normally distributed number
        //If value is less than zero or greater than six calls function again
        void GetDistance()
        {
            fDistance = fHomeNest.Normal(MeanDistance, StandardDevDistance);
            if ((fDistance > fMaxDistance) || (fDistance < 0))
                GetDistance();
        }

        //Updates fDirection by drawing a number from VonMises distribution and adding to current direction
        void GetRandomDirection()
        {
            fDirection += GetVonMisesAngle(Kappa);
        }

        //Updates fDirection by drawing a number from VonMises distribution and adding to direction of target
        void GetTargetDirection(PointF target)
        {
            float TargetDirection = (float)(Math.Atan2(target.X - fPosition.X, fPosition.Y - target.Y));

            if (TargetDirection < 0)
                TargetDirection += (float)(2 * Math.PI);

            fDirection = TargetDirection + GetVonMisesAngle(Kappa);
        }

        //Chooses new turning angle
        float GetVonMisesAngle(double kappa)
        {
            /* new turning angle chosen stochastically from distribution
            * from Best & Fisher 1979. Simulation of the von Mises Distribution. 
            * Applied Statistics 28: 152-157
            * pseudo-random angle drawn from von Mises distribution
            * the mean anlge of the distribution is given by the current heading phi
            * kappa is derived from the mean vector angle as calculated from the ant 
            * paths.*/

            double newAngle, r, z, f, c, u1, u2, u3;

            u3 = fHomeNest.Random;

            double tau = 1 + Math.Sqrt(1 + 4 * kappa * kappa);
            double rho = (tau - Math.Sqrt(2 * tau)) / (2 * kappa);
            r = (1 + rho * rho) / (2 * rho);

            while (true)
            {
                u1 = fHomeNest.Random;
                u2 = fHomeNest.Random;

                z = Math.Cos(Math.PI * u1);
                f = (1 + r * z) / (r + z);
                c = kappa * (r - f);

                if ((c * (2 - c) - u2) > 0)
                    break;

                if ((Math.Log(c / u2) + 1 - c) >= 0)
                    break;
            }

            /* acos returns values only between 0 and pi! thus, i throu out the fabs */
            newAngle = (u3 - 0.5) * Math.Acos(f);

            return (float)newAngle;
        }

        //Moves ant according to fDistance and fDirection
        void UpdatePosition()
        {
            fPosition = new PointF(fPosition.X + fDistance * ((float)Math.Sin(fDirection)),
                fPosition.Y - fDistance * ((float)Math.Cos(fDirection)));
        }

        //Returns true if there is food within aRadius of fPosition
        bool CheckForFood(float aRadius)
        {
            Food lFood = fHomeNest.World.FindFoodAt(fPosition, aRadius);

            if (lFood != null && lFood.Amount > 0)
            {
                return true;
            }

            return false;
        }

        //Returns true if there is within aRadius of fPosition
        bool CheckForNest(float aRadius)
        {
            Nest lNest = fHomeNest.World.FindNestAt(fPosition, aRadius);

            if (lNest != null)
            {
                return true;
            }

            return false;
        }

        //Exchanges food with home nest and returns false if energy crosses zero
        bool NestExchange()
        {
            bool NegativeEnergy = this.fEnergy < 0;

            if (NegativeEnergy)
            {
                this.fHomeNest.TakeFood(DepositRate);
                fEnergy += DepositRate;
            }

            if (!NegativeEnergy)
            {
                fHomeNest.DepositFood(DepositRate);
                fEnergy -= DepositRate;
            }

            if ((NegativeEnergy && (fEnergy > 0)) || (!NegativeEnergy && (fEnergy < 0)))
                return false;

            return true;
        }

        //Sets state of ant for returning home
        void SetOffHome()
        {
            fTarget = fHomeNest.Position;
            fTargetType = TargetType.Home;
            fBehaviour = Behaviour.GoingToTarget;
            fAtTarget = false;
        }

        //Takes food from FoodSource and returns false if decision to go home
        bool TakeFood(Food FoodSource)
        {
            fVisitedFoods.Add(FoodSource);

            if (((fHomeNest.Random * FeedScale) > fEnergy) && CheckForFood(Food.fRadius))
            {
                if (FoodSource.Amount > DepositRate)
                {
                    FoodSource.TakeFood(DepositRate);
                    fEnergy += DepositRate;
                    fFoundFood = true;
                    fLayTrail = true;
                    fFoundFoodEmpty = false;
                }
                else
                {
                    FoodSource.TakeFood(FoodSource.Amount);
                    fEnergy += FoodSource.Amount;
                    fFoundFood = true;
                    fLayTrail = false;
                    fFoundFoodEmpty = true;
                }
                return true;
            }
            else
            {
                return false;
            }
        }

        //Takes food from none home nest and returns false if decision to go home
        bool StealFood()
        {
            //if energy in nest is negative or zero return false (leave imediately) and lay trail if empty recruit is true
            if (fStealNest.Amount <= 0)
            {
                
                if (this.fWorld.fEmptyRecruit)
                {
                    fFoundNest = true;
                    fLayTrail = true;
                }
                else
                {
                    fLayTrail = false;
                }

                return false;
            }

            //Ants leave nests randomly or when they take the last bit of food
            if ((fHomeNest.Random * FeedScale) > fEnergy)
            {
                if (fStealNest.Amount > DepositRate)
                {
                    fStealNest.TakeNestFood(DepositRate);
                    fEnergy += DepositRate;
                    fHomeNest.StealAmount += DepositRate;
                    fFoundNest = true;
                    fFoundNestEmpty = false;
                    fLayTrail = true;
                }
                else
                {
                    fStealNest.TakeNestFood(fStealNest.Amount);
                    fEnergy += fStealNest.Amount;
                    fHomeNest.StealAmount += DepositRate;
                    fFoundNest = true;
                    fFoundNestEmpty = true;
                    if (this.fWorld.fEmptyRecruit)
                    {
                        
                        fLayTrail = true;
                    }
                    else
                    {
                        fLayTrail = false;
                    }
                }
                return true;
            }

            else
            {
                return false;
            }
        }

        //Draws single ant
        public void Draw(Graphics aGraphics)
        {
            Pen aPen;
            aPen = new Pen(Color.Black);

            PointF HeadPosition = new PointF(fPosition.X + Ant.Length * ((float)Math.Sin(Direction)), fPosition.Y - Ant.Length * ((float)Math.Cos(Direction)));
            aGraphics.DrawLine(aPen, fPosition.X, fPosition.Y, HeadPosition.X, HeadPosition.Y);
        }
    }
}
