﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Drawing;

namespace Ants
{
    public class Output
    {
        //Files
        StreamWriter Energy;
        StreamWriter Sd;
        StreamWriter FoodTime;
        StreamWriter WaitingTimes;
        StreamWriter ArrivalTimes;
        StreamWriter Info;
        
        //Properties for file names
        World fWorld;
        int fNests;
        int fFoods;
        int fAnts; 
        Colony.ForagingType fForaging;
        int fGroupSize;
        string NameForage;
        string PathofFile;

        //Constructor
        public Output(World aWorld, int aFoods, int aNests, int aAnts, Colony.ForagingType aForaging, int aGroupSize)
        {
            fWorld = aWorld;
            NameForage = aForaging.ToString();
            fForaging = aForaging;
            fNests = aNests;
            fFoods = aFoods;
            fGroupSize = aGroupSize;
            fAnts = aAnts;

            //Set folder for files to go in
            if (fForaging == Colony.ForagingType.Group)
            {
                PathofFile = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + @"\\Ants Output\\" + "Networks\\" +
                                NameForage + "\\Groups of " + aGroupSize + "\\" + aNests + " Nests\\" + aAnts + " Ants\\" + aFoods + " Foods\\";
            }
            else
            {
                PathofFile = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + @"\\Ants Output\\" + "Networks\\" +
                                NameForage + "\\" + aNests + " Nests\\" + aAnts + " Ants\\" + aFoods + " Foods\\";
            }

            //Create files and add headers
            Energy = File.CreateText(PathofFile + "Energy.txt");
            AddHeaders(Energy);

            FoodTime = File.CreateText(PathofFile + "FoodTimes.txt");
            AddHeaders(FoodTime);

            if (fNests > 1)
            {
                Sd = File.CreateText(PathofFile + "Sd.txt");
                AddHeaders(Sd);
            }

            if (fForaging == Colony.ForagingType.Group)
            {
                WaitingTimes = File.CreateText(PathofFile + "WaitTimes.txt");
                AddHeaders(WaitingTimes);
            }

            if (aForaging == Colony.ForagingType.Pheromone )
            {
                if (fFoods == 1)
                {
                    ArrivalTimes = File.CreateText(PathofFile + "ArrivalTimes.txt");
                    AddHeaders(ArrivalTimes);
                }
            }
        }

        //Run at end of each rep
        public void EndRep(World aWorld, int RepNum)
        {
            //write total colony energy
            Energy.WriteLine(fWorld.TotalFood());

            //write time colony first found food
            FoodTime.WriteLine(fWorld.FoodTime);

            //write information about food and nest locations
            Info = File.CreateText(PathofFile + "/Info/" + "Info" + RepNum + ".txt");
            AddHeaders(Info);
            Info.WriteLine("Nest positions");
            foreach (Colony lcolony in aWorld.Colonies)
            {
                foreach (Nest lnest in lcolony.Nests)
                {
                    Info.WriteLine(lnest.Position.X + ", " + lnest.Position.Y);
                }
            }
            Info.WriteLine("Food positions");
            foreach (Food lfood in aWorld.Foods)
            {
                Info.WriteLine(lfood.Position.X + ", " + lfood.Position.Y);
            }
            Info.Close();

            //write standard deviation of energy for polydomous nests
            if (fNests > 1)
            {
                Sd.WriteLine(StandDev(fWorld.NestEnergies()));
            }

            //write time ants arrive at food for pheromone with one food
            if (fForaging == Colony.ForagingType.Pheromone && fFoods == 1)
            {
                WriteArrivals(ArrivalTimes);
            }

            //write time ants spend waiting in nest in group foraging
            if (fForaging == Colony.ForagingType.Group)
            {
                foreach (int i in fWorld.WaitTimes)
                {
                    WaitingTimes.WriteLine(i);
                }
            }
        }

        //Closes all files
        public void EndAll()
        {
            Energy.Close();
            FoodTime.Close();

            if (fForaging == Colony.ForagingType.Pheromone)
            {
                if (fFoods == 1)
                {
                    ArrivalTimes.Close();
                }
            }
            if (fNests > 1)
            {
                Sd.Close();
            }
            if (fForaging == Colony.ForagingType.Group)
            {
                WaitingTimes.Close();
            }
        }

        //Draws and saves map with all colonies and food items if specified
        public void DrawMap(World aWorld, bool withPheromone, bool withAnts, bool withFood, bool withLabels, string FileName)
        {
            int SizeOfMap = 1500;
            Bitmap outpic = new Bitmap(SizeOfMap, SizeOfMap);
            Graphics outGraphics = Graphics.FromImage(outpic);
            foreach (Colony acolony in aWorld.Colonies)
            {
                ColonyDraw(outGraphics, acolony, withPheromone, withAnts, withLabels);
            }
            if (withFood)
            {
                FoodDraw(outGraphics, aWorld);
            }

            outpic.Save(PathofFile + "\\Maps\\Map" + FileName + ".png", System.Drawing.Imaging.ImageFormat.Png);
        }

        //Sets output picture size to be a bit bigger than the edges of colony
        private int MapSize(World aWorld)
        {
            double max = 0;
            foreach (Colony lcolony in aWorld.Colonies)
            {
                foreach (Nest lnest in lcolony.Nests)
                {
                    max = Math.Max(max, lnest.Position.X);
                    max = Math.Max(max, lnest.Position.Y);
                }
            }
            return (int)(max + (100 - (max % 100)));
        }

        //Calculates and outputs distance matrix of all nests
        public void WriteDistances(World aWorld, string FileName)
        {
            StreamWriter DistanceFile = File.CreateText(PathofFile + "\\Distances\\Distances" + FileName + ".txt");
            foreach (Colony lcolony in aWorld.Colonies)
            {
                DistanceFile.WriteLine("NEW COLONY" + Environment.NewLine);
                foreach (Nest lnest in lcolony.Nests)
                {
                    foreach (Nest locNest in lcolony.Nests)
                    {
                        float distance = World.FindDistance(lnest.Position, locNest.Position);
                        DistanceFile.Write(distance + "\t");
                    }
                    DistanceFile.Write(Environment.NewLine);
                }
            }
            DistanceFile.Close();
        }

        //Collects and outputs data on movement between all nests
        public void WriteMovement(World aWorld, string FileName)
        {
            StreamWriter MovementFile = File.CreateText(PathofFile + "\\Matrix\\Movement" + FileName + ".txt");
            foreach (Colony lcolony in aWorld.Colonies)
            {
                MovementFile.Write("NEW COLONY" + Environment.NewLine);
                
                for (int i = 0; i < lcolony.Nests.Count; i++)
                {
                    foreach (Nest lnest in lcolony.Nests)
                    {
                        if (lnest.NumberForOutput == i)
                        {
                            for (int j = 0; j < lcolony.Nests.Count; j++)
                            {
                                foreach (Nest locnest in lcolony.Nests)
                                {
                                    if (locnest.NumberForOutput == j)
                                    {
                                        MovementFile.Write(lnest.ArrivalsFromNests[locnest] + "\t");
                                    }
                                }
                            }
                            MovementFile.Write(Environment.NewLine);
                        }
                    }
                }
            }
            MovementFile.Close();
        }

        //Draws all foods within world
        public static void FoodDraw(Graphics aGraphics, World aWorld)
        {
            foreach (Food aFood in aWorld.Foods)
            if (aFood.Empty)
                    {
                        Pen lPen = new Pen(Color.DarkSeaGreen);
                        aGraphics.FillEllipse(lPen.Brush, (aFood.Position.X - 1), (aFood.Position.Y - 1), 3, 3);
                    }
                    else
                    {
                        Pen lPen = new Pen(Color.Brown);
                        aGraphics.DrawEllipse(lPen, (aFood.Position.X - Food.DetectRange), (aFood.Position.Y - Food.DetectRange), Food.DetectRange * 2, Food.DetectRange * 2);
                        aGraphics.FillEllipse(lPen.Brush, (aFood.Position.X - Food.fRadius), (aFood.Position.Y - Food.fRadius), Food.fRadius * 2, Food.fRadius * 2);
                    }
        }

        //Draws a colony
        public void ColonyDraw(Graphics aGraphics, Colony aColony, bool withPheromone, bool withAnts, bool withLabels)
        {
            aGraphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
            int aAge = aColony.World.Age;
            if (withPheromone)
            {
                foreach (KeyValuePair<int, Dictionary<int, List<Pheromone>>> xPair in aColony.HomePheromoneMap)
                {
                    foreach (KeyValuePair<int, List<Pheromone>> yPair in xPair.Value)
                    {
                        foreach (Pheromone pheromone in yPair.Value)
                        {
                            pheromone.Draw(aGraphics, fWorld.Age);
                        }
                    }
                }
            }

            foreach (Nest lnest in aColony.Nests)
            {
                NestDraw(aGraphics, lnest, withAnts);
                if (withLabels)
                {
                    string Label = lnest.NumberForOutput.ToString();
                    NestLabel(aGraphics, lnest, Label);
                }
            }
        }

        //Adds label to single nest
        private void NestLabel(Graphics aGraphics, Nest lnest, string number)
        {
            Font lFont = new Font("Ariel", 20);
            Brush lBrush = new SolidBrush(System.Drawing.Color.Black);
            PointF labelPosition = new PointF(lnest.Position.X + 20, lnest.Position.Y);
            aGraphics.DrawString(number, lFont, lBrush, labelPosition);
        }

        //Draws single nest and all ants within that nest if specified
        private void NestDraw(Graphics aGraphics, Nest lnest, bool withAnts)
        {
            Pen lPen = new Pen(Color.Goldenrod);
            aGraphics.DrawEllipse(lPen, (lnest.Position.X - Nest.DetectRange), (lnest.Position.Y - Nest.DetectRange), Nest.DetectRange * 2, Nest.DetectRange * 2);
            aGraphics.FillEllipse(lPen.Brush, (lnest.Position.X - Nest.fRadius), (lnest.Position.Y - Nest.fRadius), Nest.fRadius * 2, Nest.fRadius * 2);
            if (withAnts)
            {
                foreach (Ant lant in lnest.fAnts)
                {
                    lant.Draw(aGraphics);
                }
            }
        }

        //Adds header including info about number of food/nests/ants and foraging
        private void AddHeaders(StreamWriter FileStream)
        {
            FileStream.WriteLine("Food:" + fFoods + "\tNests:" + fNests + "\tForaging:" + NameForage + "\tAnts:" + fAnts + "\tGroups of:" + fGroupSize);
        }

        //Writes arrival times to file
        private void WriteArrivals(StreamWriter FileStream)
        {
            FileStream.WriteLine("ENDREP");
            if (fWorld.ArrivalTimes.Count != 0)
            {
                foreach (int i in fWorld.ArrivalTimes)
                {
                    FileStream.WriteLine(i);
                }
            }
        }

        //Calculates standard deviation of input numbers
        double StandDev(List<double> numbers)
        {
            double SumOfSqrs = 0;
            double avg = numbers.Average();
            for (int i = 0; i < numbers.Count; i++)
            {
                SumOfSqrs += Math.Pow(((double)numbers[i] - avg), 2);
            }
            double n = (double)numbers.Count;
            return Math.Sqrt(SumOfSqrs / (n - 1));
        }

    }
}
