using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Ants
{
    public class Pheromone
    {
        public enum Type
        {
            GoingHome,
            FoundFood
        }

        public const int DropInterval = 1;
        public const int MaxDrawAlpha = 6;

        public const int SensingDistance = 10;
        public const float MinimumDetection = 0.05F;

        private const int Density = 5;

        private PointF fPosition;
        private float fCreated;
        private Type fType;
        private float fInitialAmount;

        private static Pen fHomePen = new Pen(Color.FromArgb(Density, Color.BlueViolet));
        private static Pen fFoodPen = new Pen(Color.FromArgb(Density, Color.Olive));

        public PointF Position
        {
            get { return fPosition; }
        }

        //Constructor
        public Pheromone(PointF aPosition, float aInitialAmount, float aCreated, Type aType)
        {
            fPosition = aPosition;
            fCreated = aCreated;
            fType = aType;
            fInitialAmount = aInitialAmount;
        }

        //Returns amount after decay. Zero if less than MinimumDetection
        public float GetAmount(float aWorldAge)
        {
            float DecayAmount = (float)Math.Max(0, fInitialAmount * Math.Exp((aWorldAge - fCreated)*-0.005));
            return DecayAmount < Pheromone.MinimumDetection ? 0 : DecayAmount;
        }

        public bool Expired(float aAge)
        {
            float amount = GetAmount(aAge);

            if (amount == 0)
                return true;

            return false;
        }

        //Draws single pheromone point
        public void Draw(Graphics aGraphics, float aAge)
        {
            int DrawDensity = 5;
            int MaxDrawAlpha = 10;

            Pen lPen = new Pen(Color.FromArgb(DrawDensity, Color.BlueViolet));
            lPen.Color = Color.FromArgb((int)(Math.Min(1, GetAmount(aAge)) * MaxDrawAlpha), lPen.Color);
            aGraphics.FillEllipse(lPen.Brush, (Position.X - (SensingDistance / 2F)), (Position.Y - (SensingDistance / 2F)), SensingDistance, SensingDistance);
        }

    }
}
