using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.IO;
using System.Drawing;
using System.Text;

namespace Ants
{
    static class Program
    {
        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main(string[] args)
        {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);

            //create parser
            CMDLineParser parser = new CMDLineParser();

            //Add options for input
            CMDLineParser.NumberOption NumCol = parser.AddIntParameter("-c",
                                     "Number of colonies", true);
            CMDLineParser.NumberOption NumNest = parser.AddIntParameter("-n",
                                    "Number of nests per colony", true);
            CMDLineParser.NumberOption NumAnt = parser.AddIntParameter("-a",
                                    "Number of ants per nest", true);
            CMDLineParser.NumberOption NumForager = parser.AddIntParameter("-f",
                                    "Number of foragers per nest", true);
            CMDLineParser.NumberOption NumTransporter = parser.AddIntParameter("-T",
                                    "Number of transporters per nest", true);
            CMDLineParser.NumberOption NumReps = parser.AddIntParameter("-r",
                                    "Number of replicates", true);
            CMDLineParser.NumberOption NumGroup = parser.AddIntParameter("-g",
                                    "Groups size for foraging", false);
            CMDLineParser.NumberOption ForageType = parser.AddIntParameter("-t",
                                    "Type of foraging: \n\t0 = individual\n\t1 = group\n\t2 = pheromone", false);
            CMDLineParser.Option NestFile = parser.AddStringParameter("-I",
                                    "Input file containing nest locations", false);
            try
            {
                //parse the command line
                parser.Parse(args);
            }
            catch (CMDLineParser.CMDLineParserException ex)
            {
                //show available options
                System.Windows.Forms.MessageBox.Show("Error: " + ex.Message);
                System.Windows.Forms.MessageBox.Show(parser.HelpMessage());
                return;
            }

            //replace argument list with remaining arguments
            args = parser.RemainingArgs();

            //Get values 
            int aNests = (int)NumNest.Value;
            int aReps = (int)NumReps.Value;
            int aColonies = (int)NumCol.Value;
            int aAnts = (int)NumAnt.Value;
            int aForagers = (int)NumForager.Value;
            int aTransporters = (int)NumTransporter.Value;
            int aGroupSize;
            Colony.ForagingType aForagingType;
            string aNestFile = (string)NestFile.Value;

            //if group size is not specified set to zero
            if (NumGroup.Value != null)
            {
                aGroupSize = (int)NumGroup.Value;
            }
            else
            {
                aGroupSize = 0;
            }

            //Get foraging type from int input
            switch ((int)ForageType.Value)
            {
                case 0:
                    {
                        aForagingType = Colony.ForagingType.Individual;
                    }
                    break;
                case 1:
                    {
                        aForagingType = Colony.ForagingType.Group;
                    }
                    break;
                case 2:
                    {
                        aForagingType = Colony.ForagingType.Pheromone;
                    }
                    break;
                default:
                    {
                        aForagingType = Colony.ForagingType.Individual;
                    }
                    break;
            }

            //Fixed values
            int aRunLength = 50000;
            int aTotalFood = 1000000;

            //program will loop for each food number in this list
            List<int> aFoods = new List<int>();
            aFoods.Add(3);

            foreach (int aFood in aFoods)
            {
                //Create world
                World world1 = new World(aTotalFood, aFood, aColonies, aNests, aAnts, aForagers, aTransporters, aForagingType, aNestFile, aGroupSize);
                world1.RecordPaths = true;
                //Create output
                Output fOutput = new Output(world1, aFood, aNests, aAnts, aForagingType, aGroupSize);
                if (aNestFile != null)
                {
                    fOutput.WriteDistances(world1, "");
                }

                for (int i = 0; i < aReps; ++i)
                {
                    RunRep(world1, aRunLength, fOutput, aNestFile, i, null);
                }
                //output at end of all reps
                fOutput.EndAll();
            }
        }

        public static void RunRep(World aWorld, int aRunLength, Output aOutput, string aNestFile, int RepNum, Graphics graphics)
        {
            for (int j = 0; j <= aRunLength; j++)
            {
                RunStep(aWorld, aOutput, j, RepNum, graphics);
            }

            //output at end of every rep
            aOutput.EndRep(aWorld, RepNum);
            string FileName = "Rep" + RepNum;
            if (aNestFile == null)
            {
                aOutput.WriteDistances(aWorld, FileName);
            }
            aWorld.Restart();
        }

        public static void RunStep(World aWorld, Output aOutput, int StepNum, int RepNum, Graphics graphics)
        {
            aWorld.Step(graphics);
            //periodic output 
            if (StepNum % 5000 == 0 && StepNum > 0)
            {
                string File1 = "Rep_" + RepNum + "Time_" + StepNum;
                aOutput.DrawMap(aWorld, true, true, true, true, File1);
                aOutput.WriteMovement(aWorld, File1);
                aWorld.ResetMovement();
            }
        }

    }
}
