using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.IO;

namespace Ants
{
    public class World
    {
        int fNumberOfFoods;
        int fAge;
        FastRandom fRandom;

        int fColonies;
        int fNests;
        int fForagers;
        int fTransporters;

        string fNestFile;
        public static Size fSize = new Size(1000, 1000);

        public double Random
        {
            get { return fRandom.NextDouble(); }
        }
        public double fTotalFood;
        public float Normal(float mean, float stdev)
        {
            return fRandom.GetNormal(mean, stdev);
        }
        public PointF RandomLocation(Size worldsize)
        {
            return new PointF((float)Random * worldsize.Width, (float)Random * worldsize.Width);
        }
        public List<Colony> Colonies = new List<Colony>();
        public List<Food> Foods = new List<Food>();
        public List<Food> TempFoods = new List<Food>();
        //INFO: Ignore phermones close to nest bool is here
        public bool IgnoreCloseToNest = false;

        public int Age
        {
            get { return fAge; }
        }

        public bool fEmptyRecruit;
        public bool RouteFidelity;

        public List<double> NestEnergies()
        {
            List<double> list1 = new List<double>();
            foreach (Colony lColony in Colonies)
            {
                foreach (Nest LNest in lColony.Nests)
                {
                    list1.Add(LNest.Amount);
                }
            }
            return list1;
        }

        public List<int> WaitTimes = new List<int>();
        public List<int> ArrivalTimes = new List<int>();

        public int FoodTime = 0;

        // Constructor
        public World(double aTotalFood, int aFoods, int aColonies, int aNests, int aForagers, int aTransporters, string aNestFile, bool aEmptyRecruit, bool aRouteFidelity)
        {
            fEmptyRecruit = aEmptyRecruit;
            RouteFidelity = aRouteFidelity;
            fRandom = new FastRandom((int)DateTime.Now.Ticks);
            fNumberOfFoods = aFoods;
            fTotalFood = aTotalFood;
            fColonies = aColonies;
            fNests = aNests;
            fForagers = aForagers;
            fTransporters = aTransporters;
            fNestFile = aNestFile;

            //Random food locations
            //for (int i = 0; i < fNumberOfFoods; i++)
           // {
            //    AddFood(RandomLocation(fSize), (float)(fTotalFood / fNumberOfFoods), false);
            //}

            //Food location for Heller
            //AddFood(new PointF(456.953F, 543.046F), (float)(fTotalFood / fNumberOfFoods), false);

            //Food locations for Holt
            AddFood(new PointF(262.609F, 740.870F), (float)(fTotalFood / fNumberOfFoods), false);
            AddFood(new PointF(433.043F, 1125.21F), (float)(fTotalFood / fNumberOfFoods), false);
            AddFood(new PointF(584.348F, 1019.130F), (float)(fTotalFood / fNumberOfFoods), false);

            //Food locations for vanWillenburg 1
            //AddFood(new PointF(349.057F, 1047.170F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(584.906F, 735.849F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(1084.906F, 584.906F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(1160.378F, 141.509F), (float)(fTotalFood / fNumberOfFoods), false);

            //Food locations for McIver 1
            //AddFood(new PointF(88.218F, -196.06F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(205.88F, 504.93F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(725.478F, 700.998F), (float)(fTotalFood / fNumberOfFoods), false);
            
            for (int i = 0; i < aColonies; i++)
            {
                AddColony(aNests, aForagers, aTransporters, aNestFile);
            }

            //create dictionary of nests
            foreach (Colony lcolony in this.Colonies)
            {
                foreach (Nest lnest in lcolony.Nests)
                {
                    foreach (Ant lant in lnest.fAnts)
                    {
                        foreach (Nest visitnest in lcolony.Nests)
                        {
                            lant.fAllVisitedNests.Add(visitnest, 0);
                        }
                    }
                }
            }

            //create dictionary of foods
            foreach (Colony lcolony in this.Colonies)
            {
                foreach (Nest lnest in lcolony.Nests)
                {
                    foreach (Ant lant in lnest.fAnts)
                    {
                        foreach (Food lfood in this.Foods)
                        {
                            lant.fAllVisitedFoods.Add(lfood, 0);
                        }
                    }
                }
            } 
        }

        public void AddColony(int aNumOfNests, int aForagingAnts, int aTransporterAnts, string aNestFile)
        {
            Colonies.Add(new Colony(aNumOfNests, aForagingAnts, aTransporterAnts, this, aNestFile));
        }

        public void AddFood(PointF aPosition, float aAmount, bool temp)
        {
            Food lFood = new Food(aPosition, aAmount);
            if (!temp)
            {
                Foods.Add(lFood);
            }
            else
            {
                TempFoods.Add(lFood);
            }
        }

        //Check for food within aRadius of aPoint
        public Food FindFoodAt(PointF aPoint, float aRadius)
        {
            foreach (Food lFood in Foods)
            {
                if (lFood.Contains(aPoint, aRadius) && !(lFood.Empty))
                    return lFood;
            }
            
            foreach (Food lFood in TempFoods)
            {
                if (lFood.Contains(aPoint, aRadius) && !(lFood.Empty))
                    return lFood;
            }

            return null;
        }

        //Check for nests within aRadius of aPoint
        public Nest FindNestAt(PointF aPoint, float aRadius)
        {
            foreach (Colony lColony in Colonies)
            {
                foreach (Nest lNest in lColony.Nests)
                {
                    if (lNest.Contains(aPoint, aRadius))
                        return lNest;
                }
            }
            return null;
        }

        //Returns straight line distance between two points
        public static float FindDistance(PointF aPoint, PointF bPoint)
        {
            float lXDiff = aPoint.X - bPoint.X;
            float lYDiff = aPoint.Y - bPoint.Y;

            float Distance = (float)Math.Sqrt(lXDiff * lXDiff + lYDiff * lYDiff);
            return Distance;
        }

        public static bool IsPointInRadius(PointF position, PointF target, float radius)
        {
            double lXDiff = position.X - target.X;
            double lYDiff = position.Y - target.Y;

            if ((lXDiff * lXDiff + lYDiff * lYDiff) <= radius * radius)
                return true;

            return false;
        }

        //Randomly reorders ListofStuff using Fisher-Yates shuffle
        public void FYShuffleList<E>(List<E> ListofStuff)
        {
            int n = ListofStuff.Count;

            while (n > 0)
            {
                int k = fRandom.Next(n);
                E temp = ListofStuff[k];
                n--;

                temp = ListofStuff[n];
                ListofStuff[n] = ListofStuff[k];
                ListofStuff[k] = temp;
            }

        }

        //Loops through all colonies and increases world age
        public void Step(Graphics graphics)
        {
            foreach (Colony lColony in Colonies)
            {
                lColony.Step(graphics);
            }
            ++fAge;
        }

        //Sum of energy accross all nests in all colonies
        public float TotalFood()
        {
            float total = 0;
            foreach (Colony lColony in Colonies)
            {
                foreach (Nest lNest in lColony.Nests)
                {
                    total += lNest.Amount;
                }
            }
            return total;
        }

        //Resets all relevant parameters. Respawns food.
        public void Restart()
        {
            fAge = 0;
            FoodTime = 0;
            Colonies.Clear();
            Foods.Clear();
            TempFoods.Clear();
            ArrivalTimes.Clear();
            WaitTimes.Clear();

            //Random food locations
            //for (int i = 0; i < fNumberOfFoods; i++)
            //{
            //    AddFood(RandomLocation(fSize), (float)(fTotalFood / fNumberOfFoods), false);
            //}

            //Food location for Heller
            //AddFood(new PointF(456.953F, 543.046F), (float)(fTotalFood / fNumberOfFoods), false);

            //Food locations for Holt
            AddFood(new PointF(262.609F, 740.870F), (float)(fTotalFood / fNumberOfFoods), false);
            AddFood(new PointF(433.043F, 1125.21F), (float)(fTotalFood / fNumberOfFoods), false);
            AddFood(new PointF(584.348F, 1019.130F), (float)(fTotalFood / fNumberOfFoods), false);

            //Food locations for vanWillenburg 1
            //AddFood(new PointF(349.057F, 1047.170F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(584.906F, 735.849F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(1084.906F, 584.906F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(1160.378F, 141.509F), (float)(fTotalFood / fNumberOfFoods), false);
            
            //Food locations for McIver 1
            //AddFood(new PointF(88.218F, -196.06F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(205.88F, 504.93F), (float)(fTotalFood / fNumberOfFoods), false);
            //AddFood(new PointF(725.478F, 700.998F), (float)(fTotalFood / fNumberOfFoods), false);

            for (int i = 0; i < fColonies; i++)
            {
                AddColony(fNests, fForagers, fTransporters, fNestFile);
            }

            //create dictionary of nests
            foreach (Colony lcolony in this.Colonies)
            {
                foreach (Nest lnest in lcolony.Nests)
                {
                    foreach (Ant lant in lnest.fAnts)
                    {
                        foreach (Nest visitnest in lcolony.Nests)
                        {
                            lant.fAllVisitedNests.Add(visitnest, 0);
                        }
                    }
                }
            }

            //create dictionary of foods
            foreach (Colony lcolony in this.Colonies)
            {
                foreach (Nest lnest in lcolony.Nests)
                {
                    foreach (Ant lant in lnest.fAnts)
                    {
                        foreach (Food lfood in this.Foods)
                        {
                            lant.fAllVisitedFoods.Add(lfood, 0);
                        }
                    }
                }
            }
        }

        //Reads file containing list of nest locations
        public PointF[] ReadNestFile(string aNestFile, int NumNests)
        {
            PointF[] Nests = new PointF[NumNests];
            StreamReader inputline = new StreamReader(aNestFile);
            string line = inputline.ReadLine();
            int count = 0;
            while (count < NumNests)
            {
                line = inputline.ReadLine();
                string[] splitline = new string[4];
                splitline = line.Split(new Char[] { '\t' });
                Nests[count].X = (float)Convert.ToDouble(splitline[1]);
                Nests[count].Y = (float)Convert.ToDouble(splitline[2]);
                count++;
            }
            return Nests;
        }

        public static double MaxX(PointF[] List)
        {
            double max = 0;
            foreach (PointF point in List)
            {
                max = Math.Max(max, point.X);
            }
            return max;
        }

        public static double MinX(PointF[] List)
        {
            float min = 100000000000;
            foreach (PointF point in List)
            {
                min = Math.Min(min, point.X);
            }
            return min;
        }

        public static double MaxY(PointF[] List)
        {
            double max = 0;
            foreach (PointF point in List)
            {
                max = Math.Max(max, point.Y);
            }
            return max;
        }

        public static double MinY(PointF[] List)
        {
            double min = 10000000000;
            foreach (PointF point in List)
            {
                min = Math.Min(min, point.Y);
            }
            return min;
        }

        //Converts list of nests to fit into 1000x1000 size area
        public static PointF[] TransformNests(PointF[] OriginalNestList)
        {
            PointF[] ConvertedList = new PointF[OriginalNestList.Length];
            //Convert locations into to millimeters
            for (int i = 0; i < OriginalNestList.Length; i++)
            {
                ConvertedList[i].X = OriginalNestList[i].X * 1000;
                ConvertedList[i].Y = OriginalNestList[i].Y * 1000;
            }

            //Move all to set mimimum to zero
            float minX = (float)MinX(ConvertedList);
            float minY = (float)MinY(ConvertedList);
            for (int i = 0; i < OriginalNestList.Length; i++)
            {
                ConvertedList[i].X = ConvertedList[i].X - minX;
                ConvertedList[i].Y = ConvertedList[i].Y - minY;
            }

            //Scale to fit in 1000x1000
            float scaling = (float)(Math.Max(MaxX(ConvertedList), MaxY(ConvertedList)));
            for (int i = 0; i < OriginalNestList.Length; i++)
            {
                ConvertedList[i].X = (ConvertedList[i].X / scaling) * 1000;
                ConvertedList[i].Y = (ConvertedList[i].Y / scaling) * 1000;
            }
            return ConvertedList;
        }

        //Converts speed to fit scaling of nests
        public static float ConvertSpeed(PointF[] OriginalNestList, double RealSpeed)
        {
            float newspeed;
            //Transform nest locations to find scaling factor 
            PointF[] ConvertedList = new PointF[OriginalNestList.Length];
            //Convert locations into to millimeters
            for (int i = 0; i < OriginalNestList.Length; i++)
            {
                ConvertedList[i].X = OriginalNestList[i].X * 1000;
                ConvertedList[i].Y = OriginalNestList[i].Y * 1000;
            }

            //Move all to set mimimum to zero
            float minX = (float)MinX(ConvertedList);
            float minY = (float)MinY(ConvertedList);
            for (int i = 0; i < OriginalNestList.Length; i++)
            {
                ConvertedList[i].X = ConvertedList[i].X - minX;
                ConvertedList[i].Y = ConvertedList[i].Y - minY;
            }

            //Scale
            float scaling = (float)(Math.Max(MaxX(ConvertedList), MaxY(ConvertedList)));
            //TransformSpeed according to scaling
            newspeed = (float)(RealSpeed / scaling) * 1000;
            return newspeed;
        }

        //Clears counts of ants arriving at nests
        public void ResetMovement()
        {
            foreach (Colony lColony in Colonies)
            {
                foreach (Nest lNest in lColony.Nests)
                {
                    lNest.ArrivalsFromNests.Clear();
                    foreach (Nest locNest in lColony.Nests)
                    {
                        lNest.ArrivalsFromNests.Add(locNest, 0);
                    }
                }
            }
        }

        //Adds specified number of small foods
        public void AddTempFoods(int NumberFoods, int EnergyPerFood)
        {
            for (int i = 0; i < NumberFoods; i++)
            {
                AddFood(RandomLocation(fSize), (float)EnergyPerFood, true);
            }
        }

    }
}
